/*
 * $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */
#ident "$Header: ttyname_too.c 1.2 88/12/21 $"
/*
 * NAME
 *	dev_to_ttyname, ttyname_to_dev - conversion between
 *	major/minor numbers and tty names.
 *
 * SYNOPSIS
 *	char *dev_to_ttyname(dev)
 *	dev_t dev;
 *
 *	int ttyname_to_dev(name, dev)
 *	char *name;
 *	dev_t *dev;
 *
 * DESCRIPTION
 *	These functions will do conversions between device major and
 *	minor numbers and tty path names.  The path is guaranteed to
 *	be unique for major numbers between 0 and 1023 inclusive (10
 *	bits) and minor numbers between 0 and 16383 inclusive (14 bits)
 *	or 16,777,216 entries.  There is a one-to-one mapping between
 *	devices and their names.
 *
 *	dev_to_ttyname() will return a pointer to a null terminated
 *	string of the form "/dev/tty<subdir>/<basename>", where
 *	basename is a unique four character string and subdir is
 *	two characters.  The first and third characters of basename
 *	determine the directory name for subdir (ie, /dev/ttyac/abcd).
 *	Using the first and third symbols from basename causes minor
 *	numbers to rotate between 4 directories every 64 entries.  No
 *	other major numbers will use those 4 directories.
 *
 *	Not counting "/dev/", the resulting 10 character path names
 *	are suitable for use in utmp ut_line field.  The unique 4
 *	character basename can be used in the utmp ut_id field.
 *	The entire path can be constructed from the unique basename.
 *
 *	ttyname_to_dev() takes any (potiential) path to a tty device,
 *	extracts the basename and calculates the associated device.
 *	The device is then returned in 'dev'.
 *
 *	To arrive at the subdir/basename combination, the lower 10
 *	bits of the major number and lower 14 bits of the minor number
 *	are converted into four 6 bit quantities.  Each 6 bit value
 *	is converted into a unique 4 character name using base 64
 *	arithmetic.
 *
 *	Base 64 digits are arbitrarily represented by the symbols
 *	[A-Za-z0-9_+] where 'A' == 0, 'a' == 26, '0' == 52, and
 *	'+' == 63.  The number AAAA has the value 0 and AAA+ has
 *	the value 63.
 *
 * RETURN VALUE
 *	Upon successful completion, ttyname_to_dev() will return
 *	0.  If the basename of the input string is greater than
 *	four non-null characters or the basename contains symbols
 *	not found in the regular expression [A-Za-z0-9_+], a value
 *	of -1 will be returned.
 *
 *	There is no error condition for dev_to_ttyname().  It will
 *	silently truncate 'dev' to 10 bits of major and 14 bits of
 *	minor.
 *
 * EXAMPLES
 *	The following table gives an example of the directory names
 *	used for certain major numbers.
 *
 *					minor ranges
 *	major#	directory	  0 to 255	 256 to 511
 *	-----	---------	------------	------------
 *	  0	   AA		AAAA to AAA+	ABAA to ABA+
 *		   AB		AABA to AAB+	ABBA to ABB+
 *		   AC		AACA to AAC+	ABCA to ABC+
 *		   AD		AADA to AAD+	ABDA to ABD+
 *
 *	  6	   GA		GAAA to GAA+	GBAA to GBA+
 *		   GB		GABA to GAB+	GBBA to GBB+
 *		   GC		GACA to GAC+	GBCA to GBC+
 *		   GD		GADA to GAD+	GBDA to GBD+
 *
 *	The tty device names for the 16 lines of a controller with
 *	major number 6 are:
 *		major=6, minor=0	/dev/ttyGA/GAAA
 *		major=6, minor=1	/dev/ttyGA/GAAB
 *		major=6, minor=2	/dev/ttyGA/GAAC
 *		.			.
 *		.			.
 *		major=6, minor=15	/dev/ttyGA/GAAP
 *
 * BUGS
 *	These routines are designed to allow speculation, thus
 *	major numbers are not checked for validity before conversion.
 *	For instance, major=0 and major=1024 will return the same path.
 *
 *	Only the basename of 'name' is checked, so an obviously bad
 *	path may still return a valid 'dev'.  For instance, the path
 *	"/dev/hello/world/ABCD" would only look at "ABCD" which is
 *	a valid basename.
 */
/*
 * IMPLEMENTATION NOTES:
 *	Lookup tables 'digit64' and 'value64' are used for speed.
 *	They are declared as 'const' to put this static data in
 *	the text section.  The 'const' may be non-portable.
 *
 *	The returned path name is "/dev/ttyAB/CDEF", where
 *	A == C and B == E.  This will cause the directory to
 *	change every 64 entries.  In order to keep directories
 *	owned by only one major number, the entire major number
 *	must comprise C and E (therefore A and B).
 *
 *	Using the lowest 6 bits of major for C will give each major
 *	number an intuitive directory name.  To ensure the directory
 *	belongs to a major number, use the remaining 4 bits for E.
 *
 *	Using the lowest 6 bits of minor for F will cause directory
 *	entries to follow a set pattern.  To get E to change every
 *	64 entries, use the 7th and 8th bits of minor as E's lowest
 *	bits.  Put the remaining bits of minor in D.
 *	(See table in EXAMPLE.)
 *
 *	In the following diagram '+' and '|' denotes a byte boundary
 *	and 'x' denoted unused bits.  The picture for dev_t is broken
 *	(spaces) at the intervals described above.  The picture for
 *	result is broken at 6 bit (one base 64 digit) boundaries.
 *
 *       <----- major ---->   <------ minor ----->
 *      +------ --+-- ------ +-- ------ +-- ------+
 *      |xxxxxx  E|h     C   |xx    D   |El    F  |   dev_t (32 bits)
 *      +------ --+-- ------ +-- ------ +-- ------+
 *               Eh      C          D    El    F
 *
 *                  +------ --+---- ----+-- ------+
 *                  |   C     |D     Eh |El    F  |   result (24 bits)
 *                  +------ --+---- ----+-- ------+
 *
 *
 *	The 64 entries per directory (plus "." and "..") take up about
 *	1500 bytes, slightly larger than one frag.  Therefore, most
 *	entries in the same logical group can be cached with one
 *	directory read.
 */
#include <sys/types.h>		/* for dev_t */
#include <stdio.h>

#define base	64;
/*
 *	The array 'digit64[]' defines the radix-64 "digits".  Conversion
 *	from decimal is done by simply indexing into this array.  By
 *	using only 6 bits, the index will never be out of range!
 */
const
unsigned
char digit64[] =
{
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
	'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
	'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '_', '+',
};

/*
 *	The array 'value64[]' is a subset of the US ASCII character set.
 *	This sparce array contains the values for each radix-64 digit
 *	and -1 for invalid digits.  Conversion from radix-64 to decimal
 *	is done by subtracting the ascii value of the first element of
 *	this table ('+') and if the result is a positive value within
 *	range of the table, use result as an index.
 */
const signed
char value64[] =
{
	63, -1, -1, -1, -1,			/* + , - . /         */
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, /* 0 1 2 3 4 5 6 7 8 9 */
	-1, -1, -1, -1, -1, -1, -1,		/* : ; < = > ? @     */
	 0,  1,  2,  3,  4,  5,  6,  7,  8,	/* A B C D E F G H I */
	 9, 10, 11, 12, 13, 14, 15, 16, 17,	/* J K L M N O P Q R */
	18, 19, 20, 21, 22, 23, 24, 25,		/* S T U V W X Y Z   */
	-1, -1, -1, -1, 62, -1,			/* [ \ ] ^ _ `       */
	26, 27, 28, 29, 30, 31, 32, 33, 34,	/* a b c d e f g h i */
	35, 36, 37, 38, 39, 40, 41, 42, 43,	/* j k l m n o p q r */
	44, 45, 46, 47, 48, 49, 50, 51,		/* s t u v w x y z   */
};

/*
 *	The following macros are used to convert a (32 bit) dev_t
 *	into four 6 bit (base 64) values.
 */
#define C_MUNGE(d)	(((d) & 0x3f0000) << 2)
#define D_MUNGE(d)	(((d) & 0x3f00) << 4)
#define E_MUNGE(d)	(((d) & 0x3c00000) >> 14)
#define F_MUNGE(d)	((d) & 0xff)
#define SQUISH(d)	(C_MUNGE(d) | D_MUNGE(d) | E_MUNGE(d) | F_MUNGE(d))

/*
 *	The following macros are used to convert four 6 bit (base 64)
 *	values into a (32 bit) dev_t.
 */
#define C_UMUNGE(r)	(((r) & 0xfc0000) >> 2)
#define D_UMUNGE(r)	(((r) & 0x3f000) >> 4)
#define E_UMUNGE(r)	(((r) & 0xf00) << 14)
#define F_UMUNGE(r)	((r) & 0xff)
#define USQUISH(r)	(C_UMUNGE(r) | D_UMUNGE(r) | E_UMUNGE(r) | F_UMUNGE(r))

char *
dev_to_ttyname(dev)
	dev_t dev;
{
	register int i, div;
	char *ans = "/dev/ttyAA/AAAA";
#define A_END	14
#define A_DIR	 8

	div = SQUISH(dev);

#ifdef DEBUG
	fprintf(stderr, "%d = 0x%x = ", div, div);
#endif /* DEBUG */

	/*
	 * Static data will be changed with each invocation, so
	 * always calculate four places.
	 */
	for(i = A_END; i > A_END-4; i--) {
		ans[i] = digit64[div & 0x3f];	/* div % (int)base; */
		div = (unsigned)div >> 6;	/* div / (int)base; */
	}

	ans[A_DIR] = ans[A_END - 3];
	ans[A_DIR + 1] = ans[A_END - 1];
	return(ans);
}

ttyname_to_dev(name, dev)
	char *name;
	dev_t *dev;
{
	register char *namep;
	register int i, value = 0;
	extern char *strrchr();
	
	/* Get basename */
	if ((namep = strrchr(name, '/')) != NULL)
		namep++;
	else
		namep = name;

#ifdef DEBUG
	fprintf(stderr, "name = '%s', namep = '%s'\n", name, namep);
#endif /* DEBUG */

	if (! *namep || strlen(namep) > 4)
		return(-1);

	for ( ; *namep; namep++) {
		if ((i = basetod(*namep)) < 0)
			return(-1);
		value = (value << 6) | i;
	}

	*dev = USQUISH(value);

#ifdef DEBUG
	fprintf(stderr, "%s: %d = 0x%x: ", name, value, value);
	fprintf(stderr, "dev = %d\n", *dev);
#endif /* DEBUG */

	return(0);
}

static
basetod(c)
	register char c;
{
#ifdef DEBUG
	fprintf(stderr, "c = '%c'\n", c);
#endif /* DEBUG */
	c -= (char)'+';
	if ((int)c < 0 || (int)c >= sizeof(value64))
		return(-1);
	return(value64[c]);
}

#ifdef TEST
#include <sys/sysmacros.h>	/* for major(), minor() and makedev() */

main(argc, argv)
	int argc;
	char *argv[];
{
	dev_t dev;

	switch (argc) {
	case 2:
		if (ttyname_to_dev(argv[1], &dev) == -1) {
			fprintf(stderr, "error: %s: invalid name\n", argv[1]);
			return(1);
		}
		printf("major = %d, minor = %d\n", major(dev), minor(dev));
		break;
	case 3:
		dev = makedev((short)atoi(argv[1]), (short)atoi(argv[2]));
		printf("%s\n", dev_to_ttyname(dev));
		break;
	default:
		fprintf(stderr, "usage: %s <basename>\n", argv[0]);
		fprintf(stderr, "usage: %s <major> <minor>\n", argv[0]);
		return(1);
	}
	return(0);
}
#endif /* TEST */
