/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: vm_text.c 1.26 90/03/26 $"

/*
 * vm_text.c
 *	Management of shared-text sements.
 *
 * Shared-texts are really a sub-case of mapped files and are implemented
 * here as such.  Some of the old interfaces are kept for convenience of
 * calling code (eg, xrele()).
 *
 * The implementation assumes VREG vnodes are being mapped, and the
 * implementation is file-system independent.  This can be extended to
 * file-system dependent mapping as necessary by deriving the map
 * allocator and mapping function from the vfs identified in the vnode,
 * and modifying xrele(), xflush(), xumount() accordingly.
 *
 * The text of an executable binary is mapped read only, not copy-on-ref.
 * Data is mapped read-write, copy-on-ref.
 * Invalid addresses at zero are implemented by zapping the pte's.
 * Zero address space at zero is implemented by an internal mapping function.
 * Any/all of these mappings may be changed by calls to mmap() or munmap().
 */

/* $Log:	vm_text.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/seg.h"
#include "sys/vnode.h"
#include "sys/vfs.h"
#include "sys/vm.h"
#include "sys/cmn_err.h"

#include "sys/pte.h"
#include "sys/intctl.h"

#define	MM_TEXT		0			/* text u_mmap[] index */
#define	MM_DATA		1			/* data u_mmap[] index */
#define	MM_ZERO		2			/* 0@0 u_mmap[] index */

static	struct	mmap	text_mmap;		/* prototype text mmap entry */
static	struct	mmap	data_mmap;		/* prototype data mmap entry */

extern	struct	mapops	mmap_reg;		/* VREG mapping functions */

/*
 * xinit()
 *	Init text things.
 *
 * Init prototype struct mmap's for xinitpt().
 */

xinit()
{
	/*
	 * Init prototype text mmap entry.
	 * MMT_TEXT allows chgprot() to fuss with the pages in traced process.
	 */

	text_mmap.mm_off = 0;			/* 1st mapped offset */
	text_mmap.mm_1stpg = 0;			/* 1st vpn in process */
	text_mmap.mm_ops = &mmap_reg;		/* mapping operations */
	text_mmap.mm_vp = 0;			/* used only for extending mmaps */
	text_mmap.mm_prot = PROT_READ|PROT_EXEC;/* "prot" bits */
	text_mmap.mm_paged = 1;			/* paged map */
	text_mmap.mm_noio = 0;			/* IO services allowed */
	text_mmap.mm_cor = 0;			/* don't copy-on-ref the pgs */
	text_mmap.mm_type = MMT_TEXT;		/* this is a "text" */

	/*
	 * Init prototype data mmap entry.
	 */

	data_mmap = text_mmap;			/* mostly same as text */
	data_mmap.mm_cor = 1;			/* do copy-on-ref the pgs */
	data_mmap.mm_type = 0;			/* this is not a "text" */
}

/*
 * xalloc()
 *	Get handles on a mapping for text and data.
 *
 * Called from getxfile() before process has released current resources.
 *
 * Returns 0 for success, else error.
 * Assumes argument vnode is locked.
 */

xalloc(vp, tds, ds, dp, thandlep, dhandlep)
	struct	vnode	*vp;			/* vnode to map */
	register size_t tds;			/* text+data size (HW pages) */
	register size_t ds;			/* data size (HW pages) */
	ulong		dp;			/* offset (pgs) of data in file */
	ulong		*thandlep;		/* return text+data map handle*/
	ulong		*dhandlep;		/* return data map handle*/
{
	register int	val;

	/*
	 * Call to map both text + data first.
	 * This allows underlying mapper to optimize resource
	 * allocation, and speed up data-only map, below.
	 */

	val = (*mmap_reg.map_new)(vp, (ulong)0, tds, PROT_READ|PROT_EXEC,
				thandlep, (int (*)())0, 0);
	if (val > 0)
		return(val);				/* error */
	ASSERT(val == MM_PAGED, "xalloc: map type");
	/*
	 *+ While the kernel was trying to map 
	 *+ the text segment of a process into memory,
	 *+ the underlying mapping function returned a map type 
	 *+ indicating that nonpaged memory was being used.
	 */

	/*
	 * Map data again, since will have two u_mmap[]'s for this.
	 * Assume all implementations succeed for subset map if
	 * current map already exists.
	 *
	 * Allow for no data
	 */

	if (ds != 0) {
		val = (*mmap_reg.map_new)(vp, dp, ds, PROT_READ|PROT_EXEC,
					dhandlep, (int (*)())0, 0);
	}
	ASSERT(val == MM_PAGED, "xalloc: skew");
	/*
	 *+ When the kernel was trying to map 
	 *+ the data segment of a process into memory,
	 *+ the underlying mapping function returned a map type 
	 *+ indicating that nonpaged memory was being used.
	 */

	return(0);			/* success! */
}

/*
 * xalloc_bcs()
 *	Get handles on a mapping for text and data.
 *
 * Called from getxfile() before process has released current resources.
 *
 * Returns 0 for success, else error.
 * Assumes argument vnode is locked.
 */

xalloc_bcs(vp, ts, ds, dp, thandlep, dhandlep)
	struct	vnode	*vp;			/* vnode to map */
	register size_t ts;			/* text size (HW pages) */
	register size_t ds;			/* data size (HW pages) */
	ulong		dp;			/* offset (pgs) of data in file */
	ulong		*thandlep;		/* return text+data map handle*/
	ulong		*dhandlep;		/* return data map handle*/
{
	register int	val;

	/*
	 * Call to map text first.
	 */

	val = (*mmap_reg.map_new)(vp, (ulong)0, ts, PROT_READ|PROT_EXEC,
				thandlep, (int (*)())0, 0);
	if (val > 0)
		return(val);				/* error */
	ASSERT(val == MM_PAGED, "xalloc: map type");
	/*
	 *+ While the kernel was trying to map 
	 *+ the text segment of a process into memory,
	 *+ the underlying mapping function returned a map type 
	 *+ indicating that nonpaged memory was being used.
	 */

	/*
	 * Now map data, since will have two u_mmap[]'s for this.
	 *
	 * Allow for no data
	 */

	if (ds != 0) {
		val = (*mmap_reg.map_new)(vp, dp, ds, PROT_READ|PROT_EXEC,
					dhandlep, (int (*)())0, 0);
		if (val > 0)
			return(val);				/* error */

	}

	ASSERT(val == MM_PAGED, "xalloc: skew");
	/*
	 *+ When the kernel was trying to map 
	 *+ the data segment of a process into memory,
	 *+ the underlying mapping function returned a map type 
	 *+ indicating that nonpaged memory was being used.
	 */

	return(0);			/* success! */
}

/*
 * xfree()
 *	Undo what xalloc() did.
 *
 * Error unwind in getxfile() -- for some reason couldn't complete.
 * Pass "lastfd" since no real fd associated with the mapping.
 *
 * This is *NOT* used in exit -- once mapping is set up for a process,
 * standard mapped-file handling does all the work.
 */

xfree(tds, ds, dp, thandle, dhandle)
	size_t		tds;			/* text+data size (HW pages) */
	size_t		ds;			/* data size (HW pages) */
	ulong		dp;			/* offset of data in file */
	ulong		thandle;		/* text+data map handle */
	ulong		dhandle;		/* data map handle */
{
	(*mmap_reg.map_unmap)(thandle, (ulong)0, tds,
					PROT_READ|PROT_EXEC|PROT_LASTFD);
	if (ds != 0) {			/* any data? */
		(*mmap_reg.map_unmap)(dhandle, dp, ds,
					PROT_READ|PROT_EXEC|PROT_LASTFD);
	}
}

/*
 * xfree_bcs()
 *	Undo what xalloc() did.
 *
 * Error unwind in getxfile() -- for some reason couldn't complete.
 * Pass "lastfd" since no real fd associated with the mapping.
 *
 * This is *NOT* used in exit -- once mapping is set up for a process,
 * standard mapped-file handling does all the work.
 */

xfree_bcs(ts, ds, dp, thandle, dhandle)
	size_t		ts;			/* text size (HW pages) */
	size_t		ds;			/* data size (HW pages) */
	ulong		dp;			/* offset of data in file */
	ulong		thandle;		/* text+data map handle */
	ulong		dhandle;		/* data map handle */
{
	(*mmap_reg.map_unmap)(thandle, (ulong)0, ts,
					PROT_READ|PROT_EXEC|PROT_LASTFD);
	if (ds != 0) {			/* any data? */
		(*mmap_reg.map_unmap)(dhandle, dp, ds,
					PROT_READ|PROT_EXEC|PROT_LASTFD);
	}
}


/*
 * xinitpt()
 *	Set up "text" and "data" u_mmap[]'s and pte's in page-table.
 *
 * Essentially this is a special-case version of mmap().
 *
 * Called in getxfile() after map and process page table are allocated.
 *
 * Caller assures exclusive access to target pte's.
 */

xinitpt(tds, ts, ds, dp, tv, dv, thandle, dhandle)
	size_t		tds;			/* text+data size in file (HW pages) */
	size_t		ts;			/* text size (HW pages) */
	size_t		ds;			/* data size (HW pages) */
	ulong		dp;			/* offset of data in file */
	ulong		tv;			/* virtual addr of text */
	ulong		dv;			/* virtual addr of data */
	ulong		thandle;		/* text+data map handle */
	ulong		dhandle;		/* data map handle */
{
	register struct	proc	*p = u.u_procp;

	ASSERT_DEBUG(u.u_mmapmax == u.u_mmap, "xinitpt: u_mmap skew");

	/*
	 * Fill out text u_mmap[] entry and text pte's.
	 */

	u.u_mmapmax++;					/* "alloc" u_mmap[0] */
	u.u_mmap[MM_TEXT] = text_mmap;			/* most of it */
	u.u_mmap[MM_TEXT].mm_size = tds;		/* size mapped  */
	u.u_mmap[MM_TEXT].mm_pgcnt = ts - tv;		/* # pages mapped */
	u.u_mmap[MM_TEXT].mm_1stpg = tv;		/*  1st page */
	u.u_mmap[MM_TEXT].mm_handle = thandle;		/* id's mapped object */

	ptefill(dvtopte(p, tv), MAPXTOPTE(MM_TEXT)|RO, (int)(ts - tv));
	ptefill(dvtopte(p, 0), PG_INVAL, (int)tv);

	/*
	 * Fill out data u_mmap[] entry and data pte's.
	 * No init'd data (ds==ts) sets mm_pgcnt=0 (this is rare).
	 */

	u.u_mmapmax++;					/* "alloc" u_mmap[1] */
	u.u_mmap[MM_DATA] = data_mmap;			/* most of it */
	u.u_mmap[MM_DATA].mm_off = dp;			/* 1st mapped offset */
	u.u_mmap[MM_DATA].mm_1stpg = dv;		/* 1st vpn in process */
	u.u_mmap[MM_DATA].mm_size = ds;			/* size mapped  */
	u.u_mmap[MM_DATA].mm_pgcnt = ds;		/* # pages mapped */
	u.u_mmap[MM_DATA].mm_handle = dhandle;		/* id's mapped object */
	ptefill(dvtopte(p, dv), MAPXTOPTE(MM_DATA)|RW, (int)ds);
	/*
	 * Mark invalid any unused pte's found between the end of text and
	 * the start of data.
	 */
	ptefill(dvtopte(p, ts), PG_INVAL, (int)(dv-ts));
}


/*
 * xinitpt_bcs()
 *	Set up "text" and "data" u_mmap[]'s and pte's in page-table.
 *
 * Essentially this is a special-case version of mmap().
 *
 * Called in getxfile() after map and process page table are allocated.
 *
 * Caller assures exclusive access to target pte's.
 */

xinitpt_bcs(ts, ds, dp, tv, dv, thandle, dhandle)
	size_t		ts;			/* text size (HW pages) */
	size_t		ds;			/* data size (HW pages) */
	ulong		dp;			/* offset of data in file */
	ulong		tv;			/* virtual addr of text */
	ulong		dv;			/* virtual addr of data */
	ulong		thandle;		/* text+data map handle */
	ulong		dhandle;		/* data map handle */
{
	register struct	proc	*p = u.u_procp;

	ASSERT_DEBUG(u.u_mmapmax == u.u_mmap, "xinitpt: u_mmap skew");

	/*
	 * Fill out text u_mmap[] entry and text pte's.
	 */

	u.u_mmapmax++;					/* "alloc" u_mmap[0] */
	u.u_mmap[MM_TEXT] = text_mmap;			/* most of it */
	u.u_mmap[MM_TEXT].mm_size = ts;			/* size mapped  */
	u.u_mmap[MM_TEXT].mm_pgcnt = ts - tv;		/* # pages mapped */
	u.u_mmap[MM_TEXT].mm_1stpg = tv;		/*  1st page */
	u.u_mmap[MM_TEXT].mm_handle = thandle;		/* id's mapped object */

	ptefill(dvtopte(p, tv), MAPXTOPTE(MM_TEXT)|RO, (int)(ts - tv));
	ptefill(dvtopte(p, 0), PG_INVAL, (int)tv);

	/*
	 * Fill out data u_mmap[] entry and data pte's.
	 * No init'd data (ds==ts) sets mm_pgcnt=0 (this is rare).
	 */

	u.u_mmapmax++;					/* "alloc" u_mmap[1] */
	u.u_mmap[MM_DATA] = data_mmap;			/* most of it */
	u.u_mmap[MM_DATA].mm_off = dp;			/* 1st mapped offset */
	u.u_mmap[MM_DATA].mm_1stpg = dv;		/* 1st vpn in process */
	u.u_mmap[MM_DATA].mm_size = ds;			/* size mapped  */
	u.u_mmap[MM_DATA].mm_pgcnt = ds;		/* # pages mapped */
	u.u_mmap[MM_DATA].mm_handle = dhandle;		/* id's mapped object */
	ptefill(dvtopte(p, dv), MAPXTOPTE(MM_DATA)|RW, (int)ds);

	/*
	 * Mark invalid any unused pte's found between the end of text and
	 * the start of data.
	 */
	ptefill(dvtopte(p, ts), PG_INVAL, (int)(dv-ts));
}


/*
 * xumount()
 *	Free all texts (sticky and tacky) that are from a given device.
 *
 * Used in umount() to make all refs to such vnodes go away.
 */

xumount(vfsp)
	struct vfs *vfsp;
{
	mmreg_umount(vfsp);
}

/*
 * xrele()
 *	Remove a shared text segment from the text table, if possible.
 *
 * Called with locked vnode.
 *
 * Returns true if text is released, else false.
 */

bool_t
xrele(vp)
	struct vnode *vp;
{
	return (mmreg_rele(vp));
}

/*
 * xflush()
 *	Try to delete a tacky/sticky text as a means of freeing some
 *	memory and/or vnode.
 *
 * Returns true if one is released, else fails.
 */

bool_t
xflush(dosticky)
	bool_t	dosticky;
{
	return (mmreg_flush(dosticky));
}
