/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: ufs_dsort.c 1.10 89/08/07 $"

/*
 * ufs_dsort.c
 *	Seek sort for disks.  We depend on the driver
 *	which calls us using b_disksortkey as the current cylinder number.
 */

/* $Log:	ufs_dsort.c,v $
 */

/*
 * The argument dp structure holds a b_actf activity chain pointer
 * on which we keep two queues, sorted in ascending cylinder order.
 * The first queue holds those requests which are positioned after
 * the current cylinder (in the first request); the second holds
 * requests which came in after their cylinder number was passed.
 * Thus we implement a one way scan, retracting after reaching the
 * end of the drive to the first request on the second queue,
 * at which time it becomes the first queue.
 *
 * A one-way scan is natural because of the way UNIX read-ahead
 * blocks are allocated.
 *
 * It is assumed that the caller locked the queue; no locking done here.
 *
 * bufs with bp->b_flags&B_EXPRESS set are all put at the front of the queue
 * (after the current request, of course) in FIFO order.  So, a new B_EXPRESS
 * buf is placed on the queue before the first non-B_EXPRESS buffer, the
 * current requested excepted.
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/buf.h"

#define	b_cylin	b_disksortkey

#ifdef DEBUG
long b_express_count = 0;
#endif /* DEBUG */

disksort(dp, bp)
	register struct buf *dp;
	register struct buf *bp;
{
	register struct buf *ap;

#ifdef DEBUG
	if (bp->b_flags&B_EXPRESS)
		b_express_count++;
#endif /* DEBUG */

	/*
	 * If nothing on the activity queue, then
	 * we become the only thing.
	 */
	ap = dp->b_actf;
	if(ap == NULL) {
		dp->b_actf = bp;
		dp->b_actl = bp;
		bp->av_forw = NULL;
		return;
	}

	/* Put B_EXPRESS buffers FIFO at the front of the queue, right after
	 * the current request, non-B_EXPRESS buffers in normal order after
	 * all B_EXPRESS bufs.  This means always go past B_EXPRESS bufs.
	 */
	while (ap->av_forw && ap->av_forw->b_flags&B_EXPRESS)
		ap = ap->av_forw;
	if (bp->b_flags&B_EXPRESS)
		goto insert;

	/*
	 * If we lie after the first (currently active)
	 * request, then we must locate the second request list
	 * and add ourselves to it.
	 */
	if (bp->b_cylin < ap->b_cylin) {
		while (ap->av_forw) {
			/*
			 * Check for an ``inversion'' in the
			 * normally ascending cylinder numbers,
			 * indicating the start of the second request list.
			 */
			if (ap->av_forw->b_cylin < ap->b_cylin) {
				/*
				 * Search the second request list
				 * for the first request at a larger
				 * cylinder number.  We go before that;
				 * if there is no such request, we go at end.
				 */
				do {
					if (bp->b_cylin < ap->av_forw->b_cylin)
						goto insert;
					ap = ap->av_forw;
				} while (ap->av_forw);
				goto insert;		/* after last */
			}
			ap = ap->av_forw;
		}
		/*
		 * No inversions... we will go after the last, and
		 * be the first request in the second request list.
		 */
		goto insert;
	}
	/*
	 * Request is at/after the current request...
	 * sort in the first request list.
	 */
	while (ap->av_forw) {
		/*
		 * We want to go after the current request
		 * if there is an inversion after it (i.e. it is
		 * the end of the first request list), or if
		 * the next request is a larger cylinder than our request.
		 */
		if (ap->av_forw->b_cylin < ap->b_cylin ||
		    bp->b_cylin < ap->av_forw->b_cylin)
			goto insert;
		ap = ap->av_forw;
	}
	/*
	 * Neither a second list nor a larger
	 * request... we go at the end of the first list,
	 * which is the same as the end of the whole schebang.
	 */
insert:
	bp->av_forw = ap->av_forw;
	ap->av_forw = bp;
	if (ap == dp->b_actl)
		dp->b_actl = bp;
}
