/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ident	"$Header: ufs_vnodeops.c 1.94 90/08/30 $"

/*
 * ufs_vnodeops.c
 *	UNIX file-system vnode interface procedures and vnodeops structure
 *	definition.
 */

/* $Log:	ufs_vnodeops.c,v $
 */

#include "sys/types.h"
#include "sys/devsw.h"
#include "sys/sysmacros.h"
#include "sys/signal.h"
#include "sys/param.h"
#include "sys/debug.h"
#include "sys/systm.h"
#include "sys/mutex.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/timer.h"
#include "sys/proc.h"
#include "sys/file.h"
#include "sys/uio.h"
#include "sys/conf.h"
#include "sys/cmap.h"
#include "sys/ufsinode.h"
#include "sys/ufsfilsys.h"
#include "sys/ufsmount.h"
#include "sys/ufsdir.h"
#ifdef QUOTA
#include "sys/ufsquota.h"
#endif
#include "sys/vm.h"		/* For plocal */
#include "sys/var.h"
#include "sys/open.h"
#include "sys/dirent.h"
#include "sys/cmn_err.h"
#include "sys/fcntl.h"
#include "sys/flock.h"
#include "sys/pte.h"	/* For plocal */
#include "sys/plocal.h"	/* For l struct (rablock, etc) */
#include "sys/intctl.h"	/* For SPLs */
#include "sys/poll.h"
#include "sys/stropts.h"

#define ISVDEV(t) ((t == VCHR) || (t == VBLK) || (t == VFIFO))

/*
 * ufs_open()
 *	Open a UNIX file-system vnode.
 *
 * Caller passes *vpp unlocked.  It remains unlocked at exit.
 *
 * IFBLK & IFCHR are handled by spec_open().  IFIFO is handled by fifo_open().
 */

/*ARGSUSED*/
static
ufs_open(vpp, flag, cred)
	struct vnode **vpp;
	int flag;
	struct ucred *cred;
{
	struct inode *ip;
	int error;

	ip = VTOI(*vpp);

	/*
	 * Do open protocol for inode type.
	 */


	switch (ip->i_mode & IFMT) {

	case IFCHR:
	case IFBLK:
		cmn_err(CE_PANIC, "ufs_open: Attempt to open IFCHR or IFBLK");
		/*
		 *+ An attempt was made to open an IFCHR or IFBLK device via
		 *+ ufs_open().  This indicates a kernel software problem.
		 */

	case IFSOCK:
		error = ENXIO;
		break;

	case IFIFO:
		error = ENXIO;
		break;
	
	default:
		error = 0;
		break;
	}
	return (error);
}

/*
 * ufs_close()
 *	Close a UNIX file-system vnode.
 *
 * Caller passes vp unlocked.  It remains unlocked at exit.
 *
 * IFBLK & IFCHR are handled by spec_close().  IFIFO is handled by fifo_close().
 */

/*ARGSUSED*/
static
ufs_close(vp, flag, cred)
	struct	vnode	*vp;
	int	flag;
	struct	ucred	*cred;			/* not used */
{
	return (0);
}

/*
 * ufs_fdclose()
 *	vop which gets called for every file descriptor close.
 *
 * IFIFO, IFBLK & IFCHR are handled by spec_fdclose().
 *
 *	Cleanup file locks.
 */
static
ufs_fdclose(vp)
struct vnode *vp;
{
	cleanlocks(vp, USE_PID);

}

/*
 * ufs_fcntl()
 *	vop for fcntl operations.. particularly those related to
 *	record locking, and maybe for FREESP (if ever done).
 *
 *	This code does copyin/copyout.  This is strictly not a good
 *	idea for a VOP...  But, let it go for now.
 *
 *	Also, the flags passed in here are from the file descriptor.
 *	This is ALSO not right.  But for now.....
 *
 *	IFIFO, IFBLK & IFCHR are handled by spec_fcntl().
 */
static
ufs_fcntl(vp, cmd, arg, flag, offset)
	struct vnode *vp;
	int cmd, arg, flag;
	off_t offset;
{
	register struct inode *ip;

	ip = VTOI(vp);

	/*
	 * lck_fcntl() will return EINVAL if the cmd is not a file locking cmd.
	 */
	return(lck_fcntl(vp, cmd, arg, flag, offset, ip->i_size,
			 MANDLOCK(vp, ip->i_mode)));
}

/*
 * ufs_rdwr()
 *	Read or write a vnode.
 *
 * Only do "append" mode if regular file.
 *
 * Caller locks vp depending on type.  See vno_rw().
 */

/*ARGSUSED*/
static
ufs_rdwr(vp, uiop, rw, ioflag, cred)
	struct vnode *vp;
	struct uio *uiop;
	enum uio_rw rw;
	int ioflag;
	struct ucred *cred;			/* not used */
{
	register struct inode *ip;

	ip = VTOI(vp);
	if ((ioflag & IO_APPEND) && rw == UIO_WRITE && vp->v_type == VREG) {
		/*
		 * in append mode start at end of file.
		 */
		uiop->uio_offset = ip->i_size;
	}
	return (rwip(ip, uiop, rw, ioflag));
}

/*
 * rwip()
 *	Read/Write an inode, using a uio structure.
 *
 * Caller insures sufficient locking on argument inode.  This implies locked
 * inode for RW of a normal file or directory.  IFIFO, IFBLK, and IFCHR
 * files are handled by spec_rdwr(), not here.  No problem setting bits in 
 * i_flag (eg, IMARK()), since only possible harmful race is with IFREE and
 * this can't happen since calling process has a reference to the inode.
 *
 * SUN kernel clears u_error here (see below).  The only place that really
 * needs this is acct(), so we clear it there.
 */

rwip(ip, uio, rw, ioflag)
	register struct inode *ip;
	register struct uio *uio;
	enum uio_rw rw;
	int ioflag;
{
	register int n;
	register int on;
	register int type;
	struct	vnode *devvp;
	struct buf *bp;
	struct fs *fs;
	daddr_t	lbn, bn;
	int	size;
	long	bsize;
	long	osize;
	int	error = 0;
	uint	ulimit;
	unsigned long Usave = 0;
	int	starting_uio_resid;
	struct	vnode	*vp = ITOV(ip);
        int syncIO = ioflag & IO_SYNC? B_SYNC : 0;

	ASSERT(rw == UIO_READ || rw == UIO_WRITE, "rwip");
	/*
	 *+ The kernel called the function that reads/writes files, but
	 *+ the operation requested was neither a read nor a write.
	 */

	if (MANDLOCK(vp, ip->i_mode) &&
	    (error = chklock(vp, uio, rw, ioflag, ip->i_size)))
		return(error);

	if (rw == UIO_READ && uio->uio_resid == 0)
		return (0);
	if (uio->uio_offset < 0 || (uio->uio_offset + uio->uio_resid) < 0)
		return (EINVAL);
	if (rw == UIO_READ)
		IMARK(ip, IACC);
	type = ip->i_mode&IFMT;
	ASSERT_DEBUG(type != IFBLK && type != IFCHR,"rwip: invalid inode type");
	if (uio->uio_resid == 0)
		return (0);
	/*
	 * enforce system V style ulimit.
	 *	System V doesn't fail a write which CROSSES ulimit.
	 *	Instead, it allows the transfer UP TO the ulimit value.
	 *	Syscall returns number of bytes actually transferred.
	 *	Only a transfer which is entirely above ulimit is failed.
	 *
	 *	In the partial transfer case, there's some ugliness
	 *	in that uio_resid must be reduced to terminate the
	 *	transfer early.  Then, it must be reset to indicate
	 *	the number of bytes transferred.  Usave
	 *	stores the number of bytes to "short" the
	 *	transfer in case, and the value to restore into uio_resid
	 *	at the completion of the transfer.  Note, we must do
	 *	this at ALL return points for regular files returning
	 *	from write transfers (there are TWO of them).
	 */
	if (rw == UIO_WRITE && type == IFREG ) {
		ulimit = (uint)(u.u_limit << SCTRSHFT);
		if (uio->uio_offset >= ulimit)
			return(EFBIG);
		if (uio->uio_offset + uio->uio_resid >= ulimit) {
			Usave = uio->uio_offset + uio->uio_resid - ulimit;
			uio->uio_resid -= Usave;
		}
	}

	devvp = ip->i_devvp;
	fs = ip->i_fs;
	bsize = fs->fs_bsize;
	u.u_error = 0;
	starting_uio_resid = uio->uio_resid;
	do {
		lbn = uio->uio_offset / bsize;
		on = uio->uio_offset % bsize;
		n = MIN((unsigned)(bsize - on), uio->uio_resid);
		/*
		 * If inode is mapped and overlap, use "shared"
		 * pages to do the IO.  If/when do "i_lsize"
		 * might need to restrict `n' on reads.
		 */
		if ((vp->v_flag&VMAPPED)
		&&  mmreg_rw(vp, lbn, rw, on, n, uio))
			continue;
		if (rw == UIO_READ) {
			int diff; 

			diff = ip->i_size - uio->uio_offset;
			if (diff <= 0)
				return (0);
			if (diff < n)
				n = diff;
			bn = fsbtodb(fs, bmap(ip, lbn, B_READ, on+n));
			if (u.u_error)
				return (u.u_error);
		} else {
			if (n == bsize && type == IFREG)
				bn = fsbtodb(fs, bmap(ip, lbn, B_WRITE|B_NOCLR|syncIO, n));
			else
				bn = fsbtodb(fs, bmap(ip, lbn, B_WRITE|syncIO, on+n));
			/*
			 * POSIX & SVID require we write any bytes there are
			 * room for.  If we ran out of space and we were
			 * writing more than a frag, retry asking for a
			 * single frag.
			 */
			if ((u.u_error == ENOSPC) && (n > fs->fs_fsize)) {
				u.u_error = 0;
				n = fs->fs_fsize;
				bn = fsbtodb(fs, bmap(ip, lbn, B_WRITE|syncIO, on+n));
			}
			if (u.u_error || (long)bn<0) {
				/*
				 * If a partial transfer has happened,
				 * return success.  (POSIX & SVID requirement.)
				 */
				if (starting_uio_resid != uio->uio_resid)
					u.u_error = 0;
				uio->uio_resid += Usave;
				return (u.u_error);
			}
			osize = ip->i_size;
			if (uio->uio_offset + n > osize
			&&  (type == IFDIR || type == IFREG || type == IFLNK))
				ip->i_size = uio->uio_offset + n;
		}
		size = blksize(fs, ip, lbn);
		if (rw == UIO_READ) {
			if ((long)bn<0) {
				bp = geteblk((int)size);
				clrbuf(bp);
			} else if (ip->i_lastr + 1 == lbn)
				bp=breada(devvp, bn, (int)size, 
					l.rablock, l.rasize);
			else
				bp = bread(devvp, bn, (int)size);
			ip->i_lastr = lbn;
		} else {
			if (n == bsize) 
				bp = getblk(devvp, bn, size, 0);
			else
				bp = bread(devvp, bn, (int)size);
		}
		n = MIN(n, size - bp->b_resid);
		if (bp->b_flags & B_ERROR) {
			error = EIO;
			brelse(bp);
			goto bad;
		}
		u.u_error =
		    uiomove(bp->b_un.b_addr+on, n, rw, uio);
		if (rw == UIO_READ) {
			if (n + on == bsize || uio->uio_offset == ip->i_size)
				bp->b_flags |= B_AGE;
			brelse(bp);
		} else {
			if (syncIO || type == IFDIR)
				bwrite(bp);
			else if (n + on == bsize) {
				bp->b_flags |= B_AGE;
				bawrite(bp);
			} else
				bdwrite(bp);
			vp->v_flag |= VMAPSYNC;	/* fsync before map */
			IMARK(ip, IUPD|ICHG);
		}
	} while (u.u_error == 0 && uio->uio_resid > 0 && n != 0);
	if (syncIO && rw == UIO_WRITE && (ip->i_flag & (IUPD|ICHG))) {
		iupdat(ip, 1);
	}
	if (error == 0)				/* XXX */
		error = u.u_error;		/* XXX */
bad:
	if (error && type == IFREG && rw == UIO_WRITE)
		itrunc(ip, (ulong)osize);	/* in case B_NOCLR extended */
	uio->uio_resid += Usave;
	return (error);
}

/*
 * ufs_ioctl()
 *	Do an IO-control operation.
 *
 * Caller passes vp unlocked.  It remains unlocked at exit.
 *
 * The only valid inode type here is IFCHR which is actually handled by 
 * spec_ioctl().
 */


/*ARGSUSED*/
static
ufs_ioctl(vp, cmd, arg, flag, cred)
	struct vnode *vp;
	int cmd;
	caddr_t arg;
	int flag;
	struct ucred *cred;			/* not used */
{
#ifdef DEBUG
	cmn_err(CE_PANIC, "ufs_ioctl: shouldn't be called");
#endif	/* DEBUG */
	return(EINVAL);
}

/*
 * ufs_getattr()
 *	Return attributes of UNIX file-system vnode.
 *
 * Caller passes vp locked.  It remains locked at exit.
 *
 * Need to handle all inode types here including IFBLK, IFCHR, & IFIFO 
 * since spec_getattr() calls this.
 */

/*ARGSUSED*/
static
ufs_getattr(vp, vap, cred)
	struct vnode *vp;
	register struct vattr *vap;
	struct ucred *cred;			/* not used */
{
	register struct inode *ip;

	ip = VTOI(vp);
	/*
	 * Copy from inode table.
	 */
	vap->va_type = IFTOVT(ip->i_mode);
	vap->va_mode = ip->i_mode;
	vap->va_uid = ip->i_uid;
	vap->va_gid = ip->i_gid;
	vap->va_fsid = ip->i_dev;
	vap->va_nodeid = ip->i_number;
	vap->va_nlink = ip->i_nlink;
	vap->va_flags = 0;
	vap->va_size = ip->i_size;
	vap->va_atime = ip->i_atime;
	vap->va_mtime = ip->i_mtime;
	vap->va_ctime = ip->i_ctime;
	vap->va_rdev = ip->i_rdev;
	vap->va_blocks = ip->i_blocks;
	switch (ip->i_mode & IFMT) {

	case IFBLK:
		vap->va_blocksize = BLKDEV_IOSIZE;
		break;

	case IFCHR:
		vap->va_blocksize = MAXBSIZE;
		break;

	default:
		vap->va_blocksize = ip->i_fs->fs_bsize;
		break;
	}
	return (0);
}

/*
 * ufs_setattr()
 *	Set attributes of UNIX file-system vnode.
 *
 * Caller passes vp locked.  It remains locked at exit.
 *
 * Need to handle all inode types here including IFBLK, IFCHR, & IFIFO 
 * since spec_setattr() calls this.
 */

static
ufs_setattr(vp, vap, cred)
	register struct vnode *vp;
	register struct vattr *vap;	
	struct ucred *cred;
{
	register struct inode *ip;
	timeval_t timenow;
	int chtime = 0;
	int error = 0;

	/*
	 * Cannot set certain attributes.
	 */

	if (vap->va_nlink != -1 || vap->va_blocksize != -1 ||
	    vap->va_rdev != -1 || vap->va_blocks != -1 ||
	    vap->va_fsid != -1 || vap->va_nodeid != -1 ||
	    (int)vap->va_type != -1) {
		return (EINVAL);
	}

	ip = VTOI(vp);

	/*
	 * Change file access modes.  Must be owner or su.
	 */

	if (vap->va_mode != (ushort)0xffff) {
		error = OWNER(cred, ip);
		if (error)
			goto out;
		ip->i_mode &= IFMT;
		ip->i_mode |= vap->va_mode & ~IFMT;
		if (cred->cr_uid != 0) {
			if ((ip->i_mode & IFMT) != IFDIR)
				ip->i_mode &= ~ISVTX;
			if (!groupmember(ip->i_gid))
				ip->i_mode &= ~ISGID;
		}
#if	ISVTX != VSVTX
	ERROR -- implementation assumes ISVTX==VSVTX
#endif
		vp->v_flag &= ~VSVTX;
		vp->v_flag |= (ip->i_mode & ISVTX);
		vp->v_flag &= ~VMNDLCK;
		if (MANDLOCK(vp, ip->i_mode))
			vp->v_flag |= VMNDLCK;
		IMARK(ip, ICHG);
	}

	/*
	 * Change file ownership. 
	 */

	if (vap->va_uid != INVAL_UID || vap->va_gid != INVAL_GID) {
		error = chown1(ip, vap->va_uid, vap->va_gid);
		if (error)
			goto out;
	}

	/*
	 * Truncate file. Must have write permission and not be a directory.
	 * can't be a FIFO, either.
	 */

	if (vap->va_size != -1) {
		if ((ip->i_mode & IFMT) == IFDIR) {
			error = EISDIR;
			goto out;
		}
		if ((ip->i_mode & IFMT) != IFREG) {
			error = EINVAL;
			goto out;
		}
		if (iaccess(ip, IWRITE)) {
			error = u.u_error;
			goto out;
		}
		if (VN_MAPPED(vp)) {		/* VMAPPED can be stale */
			error = EBUSY;
			goto out;
		}
		if (MANDLOCK(vp, ip->i_mode)) {
			struct uio uio;

			uio.uio_offset = vap->va_size;
			uio.uio_resid = 0;
			if (error = chklock(vp, &uio, UIO_WRITE, IO_NDELAY,
						ip->i_size))
				goto out;
		}
		itrunc(ip, vap->va_size);
	}

	/*
	 * Logically grow a file.  Used in mmap() to increase size of
	 * file prior to allocation zero-fill blocks.
	 * Caller guarantees write access to file, and file
	 * is regular file (eg, not directory, FIFO, etc).
	 */

	if (vap->va_extfile != -1)
		ip->i_size = vap->va_extfile;

	/*
	 * Change file access or modified times.
	 * If the pointer to the time structure passed as argument to the
	 * syscall is null, then we must have write access to the inode
	 * (if not owner).  We know that the ptr is null since tv_sec
	 * is 0 and tv_usec is -1.
	 *
	 * See the comment in utimes() - sys/vfs_syscalls.c.
	 */

	if (vap->va_mtime.tv_sec != -1) {
		error = OWNER(cred, ip);
		chtime = 1;
		timenow = time;
		if (vap->va_mtime.tv_sec == 0 && vap->va_mtime.tv_usec == -1) {
			if (error) {
				if (iaccess(ip, IWRITE)) {
					error = u.u_error;
					goto out;
				}
				error = 0;
			}
			/*
			 * Make sure mod time is unique ala IMARK().
			 */
			if (ip->i_mtime.tv_sec == timenow.tv_sec &&
			    ip->i_mtime.tv_usec >= timenow.tv_usec) {
				timenow.tv_usec = ip->i_mtime.tv_usec + 1;
				if (timenow.tv_usec >= 1000000) {
					timenow.tv_usec = 0;
					timenow.tv_sec++;
				}
			}
			ip->i_mtime = timenow;
		} else if (error) {
			error = EPERM;
			goto out;
		} else
			ip->i_mtime = vap->va_mtime;
	}
	if (vap->va_atime.tv_sec != -1) {
		error = OWNER(cred, ip);
		if (chtime++ == 0)
			timenow = time;
		if (vap->va_atime.tv_sec == 0 && vap->va_atime.tv_usec == -1) {
			if (error) {
				if (iaccess(ip, IWRITE)) {
					error = u.u_error;
					goto out;
				}
				error = 0;
			}
			ip->i_atime = timenow;
		} else if (error) {
			error = EPERM;
			goto out;
		} else
			ip->i_atime = vap->va_atime;
	}
	if (chtime) {
		ip->i_flag |= IACC|IUPD|ICHG;
		ip->i_ctime = timenow;
	}
out:
	iupdat(ip, 1);			/* XXX should be asyn for perf */
	return (error);
}

/*
 * chown1()
 *	Perform chown operation on inode ip.
 *
 * Caller passes locked inode.
 */

static
chown1(ip, uid, gid)
	register struct inode *ip;
	uid_t uid;
        gid_t gid;
{
  	extern int chown_restricted;
#ifdef QUOTA
	register long change;
	struct mount *mp;
#endif
	if (uid == INVAL_UID)
		uid = ip->i_uid;
	if (gid == INVAL_GID)
		gid = ip->i_gid;

	/*
	 * If chown is restricted to super-user:
	 *    If:
	 *	1) not the owner of the file, or
	 *	2) trying to change the owner of the file, or
	 *	3) trying to change the group of the file to a group not in the
	 *	   process's group set,
	 *    then must be super-user.
	 * else
	 *    If:
	 *	 not the owner of the file, then must be super-user
	 */

	if (chown_restricted) {
		if ((u.u_uid != uid || uid != ip->i_uid ||
		!groupmember(gid)) && !suser())
			return(EPERM);
	} else {
		if (u.u_uid != ip->i_uid && !suser())
			return(EPERM);
	}

#ifdef QUOTA
	if (ip->i_uid == uid)	/* this just speeds things a little */
		change = 0;
	else
		change = ip->i_blocks;

	if (ip->i_dquot != NULL) {
		struct dquot *dqp;

		(void) chkdq(ip, -change, 1);
		(void) chkiq(VFSTOM(ITOV(ip)->v_vfsp), ip, ip->i_uid, 1);
		dqp = ip->i_dquot;
		ip->i_dquot = NULL;
		dqrele(dqp);
	}
#endif
	ip->i_uid = uid;
	ip->i_gid = gid;
	IMARK(ip, ICHG);
	if (u.u_uid != 0)
		ip->i_mode &= ~(ISUID|ISGID);
#ifdef QUOTA
	mp = VFSTOM(ITOV(ip)->v_vfsp);
	if (mp->m_qinod != NULL) {
		ip->i_dquot = getinoquota(ip);
		(void) chkdq(ip, change, 1);
		(void) chkiq(mp, (struct inode *)NULL, uid, 1);
		return (u.u_error);		/* should == 0 ALWAYS !! */
	}
#endif
	return (0);
}

/*
 * ufs_access()
 *	See if given access mode is allowed.
 *
 * Return 0 for access allowed, or error number.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

/*ARGSUSED*/
static
ufs_access(vp, mode, cred)
	struct vnode *vp;
	int mode;
	struct ucred *cred;	/* not used */
{
	return (iaccess(VTOI(vp), mode));
}

/*
 * ufs_readlink()
 *	Read contents of a symbolic-link.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

/*ARGSUSED*/
static
ufs_readlink(vp, uiop, cred)
	struct vnode *vp;
	struct uio *uiop;
	struct ucred *cred;			/* not used */
{
	register struct inode *ip;
	register int error, tlen, maxlen;

	if (vp->v_type != VLNK)
		return (EINVAL);
	ip = VTOI(vp);
	
	/*
	 * If conditional symbolic link, read appropriate one based on
	 * callers current "universe".
	 */
	 
	maxlen = uiop->uio_resid;
	if (ip->i_pflags & IP_CSYMLN) {
		tlen = ip->i_ucbln;
		uiop->uio_offset = 0;
	} else
		tlen = maxlen;

	uiop->uio_resid = uiop->uio_iov->iov_len = MIN(tlen, maxlen);

	error = rwip(ip, uiop, UIO_READ, 0);

	/*
	 * Restore residual count to what the caller would expect.
	 */

	uiop->uio_resid = maxlen - (MIN(tlen,maxlen) - uiop->uio_resid);
	return (error);
}

/*
 * ufs_fsync()
 *	Synch-out an inode -- insure all it's dirty blocks are written.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

/*ARGSUSED*/
static
ufs_fsync(vp, cred)
	struct vnode *vp;
	struct ucred *cred;			/* not used */
{
	vp->v_flag &= ~VMAPSYNC;		/* about to sync it out */
	syncip(VTOI(vp));
	return (0);
}

/*
 * Unix file system operations having to do with directory manipulation.
 */

/*
 * ufs_lookup()
 *	Look up a component of a path-name in a directory.
 *
 * Caller passes dvp locked.
 * On return: dvp is unlocked, *vpp is locked.
 * If error, dvp is still unlocked, *vpp is unchanged.
 */

/*ARGSUSED*/
static
ufs_lookup(dvp, nm, vpp, cred)
	struct	vnode	*dvp;
	struct	vnode	**vpp;
	char	*nm;
	struct ucred *cred;	/* not used */
{
	struct	inode	*ip;
	int	error;

	error = dirlook(VTOI(dvp), nm, &ip);	/* returns dvp unlocked */
	if (error == 0) {
		*vpp = ITOV(ip);
		/*
		 * If vnode is a device return special vnode instead
		 *	pass original vnode in locked.
		 *	specvp unlocks it always (success OR failure)
		 *	newvp (if successful) is returned locked.
		 */
		if (ISVDEV((*vpp)->v_type)) {
			struct vnode *newvp;

			error = specvp(*vpp, &newvp);
			*vpp = newvp;
		}
	}
	return (error);
}


/*
 * ufs_create()
 *	Create a new UNIX file-system file.
 *
 * Caller passes dvp unlocked.  *vpp is returned locked.
 */

static
ufs_create(dvp, nm, vap, exclusive, mode, vpp, cred)
	struct vnode *dvp;
	char *nm;
	struct vattr *vap;
	enum vcexcl exclusive;
	int mode;
	struct vnode **vpp;
	struct ucred *cred;
{
	register int error;
	struct inode *ip;

	if (cred->cr_uid != 0) 
		vap->va_mode &= ~VSVTX;
	ip = (struct inode *) 0;
	/*
	 * allow creation of directories.
	 */
	error = direnter(VTOI(dvp), nm, DE_CREATE,
			(struct inode *)0, (struct inode *)0, vap, &ip, 1);

	/*
	 * If file exists and this is a nonexclusive create,
	 * check for not directory and access permissions.
	 */

	if (error == EEXIST) {
		switch (exclusive) {
		case NONEXCL:
			if ((ip->i_mode & IFMT) == IFDIR && (mode & IWRITE)) {
				error = EISDIR;
			} else if (mode)
				error = iaccess(ip, mode);
			else
				error = 0;
			break;
		case EXCL:
			if (vap->va_type == VSOCK) {
				int	accerr;
				if (accerr = iaccess(ip, IWRITE))
					error = accerr;
			}
			break;
		}
		if (error) {
			IPUT(ip);
		} else if ((ip->i_mode&IFMT) == IFREG && vap->va_size == 0) {
			/*
			 * Truncate regular files, if required.
			 * No need to do this if error != EEXIST, since
			 * direnter() made a new inode.
			 */
			if (MANDLOCK(ITOV(ip), ip->i_mode) &&
			    (ITOV(ip)->v_filocks != NULL)) {
				IPUT(ip);
				error = EAGAIN;
			} else if (VN_MAPPED(ITOV(ip))) {
				IPUT(ip);
				error = EBUSY;
			} else {
				itrunc(ip, (ulong) 0);
			}
		}
	} 
	if (error)
		return (error);

	*vpp = ITOV(ip);
	/*
	 * If vnode is a device return special vnode instead
	 *	pass original vnode in locked.
	 *	specvp unlocks it always (success OR failure)
	 *	newvp (if successful) is returned locked.
	 */
	if (ISVDEV((*vpp)->v_type) && exclusive == NONEXCL) {
		struct vnode *newvp;

		error = specvp(*vpp, &newvp);
		*vpp = newvp;
		if (error)
			return(error);
	}
	if (MANDLOCK(*vpp, ip->i_mode))
		(*vpp)->v_flag |= VMNDLCK;
	if (vap != (struct vattr *)0)
		(void) VOP_GETATTR(*vpp, vap, cred);

	return (error);
}

/*
 * ufs_remove()
 *	Remove a file-system name from a file (unlink), given containing
 *	directory vnode and component name to remove.
 *
 * Caller passes vp unlocked.  It remains unlocked at exit.
 */

/*ARGSUSED*/
static
ufs_remove(vp, nm, cred)
	struct vnode *vp;
	char *nm;
	struct ucred *cred;			/* not used */
{
	return (dirremove(VTOI(vp), nm, (struct inode *)0, 0));
}

/*
 * ufs_link()
 *	Link a file or a directory.
 *
 * If source is a directory, must be superuser.
 *
 * Caller passes vp, tdvp unlocked.  They remain unlocked at exit.
 */

/*ARGSUSED*/
static
ufs_link(vp, tdvp, tnm, cred)
	struct vnode *vp;
	struct vnode *tdvp;
	char *tnm;
	struct ucred *cred;			/* not used */
{
	register struct inode *sip;
	register int error;
	struct vnode *realvp;

	if (VOP_REALVP(vp, &realvp) == 0) {
		vp = realvp;
	}

	sip = VTOI(vp);
	if ((sip->i_mode & IFMT) == IFDIR && !suser()) {
		return (EPERM);
	}
	error = direnter(VTOI(tdvp), tnm, DE_LINK,
	(struct inode *)0, sip, (struct vattr *)0, (struct inode **)0,0);
	return (error);
}

/*
 * ufs_rename()
 *	Rename a file or directory
 *
 * We are given the vnode and entry string of the source and the
 * vnode and entry string of the place we want to move the source to
 * (the target). The essential operation is:
 *	unlink(target);
 *	link(source, target);
 *	unlink(source);
 * but "atomically". Can't do full commit without saving state in the inode
 * on disk, which isn't feasible at this time. Best we can do is always
 * guarantee that the TARGET exists.
 *
 * Caller passes sdvp, tdvp unlocked.  They remain unlocked at exit.
 */

/*ARGSUSED*/
static
ufs_rename(sdvp, snm, tdvp, tnm, cred)
	struct vnode *sdvp;		/* old (source) parent vnode */
	char *snm;			/* old (source) entry name */
	struct vnode *tdvp;		/* new (target) parent vnode */
	char *tnm;			/* new (target) entry name */
	struct ucred *cred;		/* not used */
{
	struct inode *sip;		/* source inode */
	register struct inode *sdp;	/* old (source) parent inode */
	register struct inode *tdp;	/* new (target) parent inode */
	register int error;

	sdp = VTOI(sdvp);
	tdp = VTOI(tdvp);

	/*
	 * Make sure we can delete the source entry.
	 */

	error = iaccess(sdp, IWRITE);
	if (error)
		return (error);


	/*
	 * Look up inode of file we're supposed to rename.
	 */

	ILOCK(sdp);				/* dirlook expects this */
	error = dirlook(sdp, snm, &sip);	/* returns sdp unlocked */
	if (error)
		return (error);

	/*
	 * Check for renaming '.' or '..' or alias of '.'
	 */

	if (strcmp(snm, ".") == 0 || strcmp(snm, "..") == 0 || sdp == sip) {
		error = EINVAL;
	        IUNLOCK(sip);		/* unlock inode, locked in dirlook */
		goto out;
	}
	
	/*
	 * Honor the SYSV sticky-bit-on-directory semantics.
	 * Do this here (eventhough we will do it again later in dirremove)
	 * to ATTEMPT to avoid creating the link in the target dir 
	 * and then finding that we don't have permission to remove the
	 * link in the source dir.
	 */
        if (sdp->i_mode & ISVTX && u.u_uid != 0 && sdp->i_uid != u.u_uid &&
                sip->i_uid != u.u_uid && iaccess(sip, IWRITE)) {
                error = EACCES;
	        IUNLOCK(sip);		/* unlock inode, locked in dirlook */
                goto out;
	}
        IUNLOCK(sip);		/* unlock inode, locked in dirlook */

	/*
	 * Link source to the target.
	 */

	error = direnter(tdp, tnm, DE_RENAME,
		sdp, sip, (struct vattr *)0, (struct inode **)0,0);
	if (error) {
		if (error != ESAME)
			goto out;
		/*
		 * ESAME ==> this is an attempt to rename one link of a file
		 * to another, or rename a file to itself.  POSIX considers
		 * this to be a NOP.  See section 5.5.3.2 of the POSIX
		 * 1003.1-1988 standard.
		 */
		error = 0;
		goto out;
	}

	/*
	 * Unlink the source entry.  dirremove() checks that the entry
	 * still reflects sip, and returns an error if it doesn't.
	 * If the entry has changed just forget about it. 
	 * Release the source inode.
	 */

	error = dirremove(sdp, snm, sip, 0);
	if (error == ENOENT)
		error = 0;
out:
	IRELE(sip);
	return (error);
}

/*
 * ufs_mkdir()
 *	Create a new UNIX file-system directory.
 *
 * Caller passes dvp unlocked.  Returns *vpp locked (if !error).
 */

/*ARGSUSED*/
static
ufs_mkdir(dvp, nm, vap, vpp, cred)
	struct vnode *dvp;
	char *nm;
	register struct vattr *vap;
	struct vnode **vpp;
	struct ucred *cred;		/* not used */
{
	struct inode *ip;
	register int error;

	error = direnter(VTOI(dvp), nm, DE_CREATE,
			(struct inode *)0, (struct inode *)0, vap, &ip,0);
	if (error == 0)
		*vpp = ITOV(ip);
	else if (error == EEXIST) {
		IPUT(ip);
	}
	return (error);
}

/*
 * ufs_rmdir()
 *	Remove a directory, given parent directory vnode and name to remove.
 *
 * Caller passes vp unlocked.  It remains unlocked at exit.
 */

/*ARGSUSED*/
static
ufs_rmdir(vp, nm, cred)
	struct vnode *vp;
	char *nm;
	struct ucred *cred;	/* not used */
{
	return (dirremove(VTOI(vp), nm, (struct inode *)0, 1));
}

/*
 * ufs_readdir()
 *	Read contents of a directory in a file-system independent format.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

/*ARGSUSED*/
static
ufs_readdir(vp, uiop, cred)
	struct vnode *vp;
	register struct uio *uiop;
	struct ucred *cred;			/* not used */
{
	register struct iovec *iovp;
	register unsigned count;
	register unsigned adjresid;
	register int ret;

	iovp = uiop->uio_iov;
	count = iovp->iov_len;
	if (uiop->uio_iovcnt != 1 || count < DIRBLKSIZ ||
	    (uiop->uio_offset & (DIRBLKSIZ -1)))
		return (EINVAL);
	count &= ~(DIRBLKSIZ - 1);
	adjresid = iovp->iov_len - count;
	uiop->uio_resid -= adjresid;
	iovp->iov_len = count;
	ret = rwip(VTOI(vp), uiop, UIO_READ, 0);
	uiop->uio_resid += adjresid;
	return (ret);
}

#ifdef DEBUG
int getdents_debug[20];
int getdents_lvl = 0;
#endif /* DEBUG */
/*
 * ufs_getdents()
 *	Read a directory and translate to system V getdents format.
 *
 * Caller insures sufficient locking on argument inode. 
 */

/* ARGSUSED */
static
ufs_getdents(vp, uio, cred)
	struct vnode *vp;
	struct uio *uio;
	struct ucred *cred;	/* not used */
{
	struct inode *ip = VTOI(vp);
	int n, on;
	struct	vnode *devvp;
	struct buf *bp;
	struct fs *fs;
	daddr_t	lbn, bn;
	int	size;
	int	len;
	long	bsize;
	int	error = 0;
	struct buf	*ebp;
	off_t	diroff;
	struct dirent	*dntp;
	struct direct	*dctp;
	off_t	tresid;

	ASSERT((ip->i_mode&IFMT) == IFDIR, "ufs_getdents type");
	/*
	 *+ The kernel called the filesystem-specific directory read
	 *+ function with an inode that was not a directory.
	 */

	/*
	 * If the requested number of bytes is less than the minimum
	 * number of bytes necessary to hold a directory entry, then
	 * return EINVAL.
	 */ 
	if (uio->uio_resid < DNTSIZ(1)) {
#ifdef DEBUG
		getdents_debug[0]++;
#endif /* DEBUG */
		return (EINVAL);
	}
	diroff = uio->uio_offset;
	if (diroff < 0) {
#ifdef DEBUG
		getdents_debug[1]++;
#endif /* DEBUG */
		return (EINVAL);
	}
	if (diroff >= ip->i_size) {
#ifdef DEBUG
		getdents_debug[14]++;
#endif /* DEBUG */
		return(0);
	}
	IMARK(ip, IACC);
	devvp = ip->i_devvp;
	fs = ip->i_fs;
	bsize = fs->fs_bsize;
	ebp = geteblk((int)bsize);
	/*
	 * can geteblk() fail??
	 */
	if (ebp->b_flags & B_ERROR) {
		error = EIO;
		brelse(ebp);
		return(error);
	}
	dntp = (struct dirent *)ebp->b_un.b_addr;
#ifdef DEBUG
	getdents_debug[12]++;
	if (getdents_lvl)
		printf("diroff %d bsize %d\n", diroff, bsize);
#endif /* DEBUG */
	u.u_error = 0;
	tresid = uio->uio_resid;
	do {

#ifdef DEBUG
		getdents_debug[2]++;
#endif /* DEBUG */
		lbn = diroff / bsize;
		on = diroff % bsize;

		n = ip->i_size - diroff;
		if (n <= 0) {
#ifdef DEBUG
			getdents_debug[3]++;
#endif /* DEBUG */
			goto finish;
		}
		n = MIN((unsigned)(bsize - on), (unsigned)n);
		bn = fsbtodb(fs, bmap(ip, lbn, B_READ, on+n));
		if (u.u_error)
			goto finish;
		size = blksize(fs, ip, lbn);
		if (ip->i_lastr + 1 == lbn) {
			bp=breada(devvp, bn, (int)size, 
				l.rablock, l.rasize);
#ifdef DEBUG
			getdents_debug[4]++;
#endif /* DEBUG */
		} else {
			bp = bread(devvp, bn, (int)size);
#ifdef DEBUG
			getdents_debug[5]++;
#endif /* DEBUG */
		}
		IMARK(ip, IACC);
		ip->i_lastr = lbn;
		n = MIN(n, size - bp->b_resid);
		if (bp->b_flags & B_ERROR) {
			error = EIO;
			brelse(bp);
			goto finish;
		}
#ifdef DEBUG
		if (getdents_lvl)
			printf("bread lbn %d on %d n %d diroff %d\n", lbn, on, n, diroff);
#endif /* DEBUG */

		/*
		 * start on the previous DIRBLKSIZE boundary, and skip through
		 * the block until >= on 
		 */
		dctp = (struct direct *)(bp->b_un.b_addr +
				(on & ~(DIRBLKSIZ - 1)));
		diroff &= ~(DIRBLKSIZ - 1);
		while (dctp < (struct direct *)(bp->b_un.b_addr + on)) {
			if (len = dirmangled(ip, dctp, diroff)) {
				diroff += len;
				dctp = (struct direct *)
					((int)dctp + len);
#ifdef DEBUG
				getdents_debug[15]++;
#endif /* DEBUG */
				continue;
			}
			diroff += dctp->d_reclen;
			dctp = (struct direct *)((int)dctp + dctp->d_reclen);
#ifdef DEBUG
			getdents_debug[6]++;
			if (getdents_lvl)
				printf("skip diroff %d\n", diroff);
#endif /* DEBUG */
		}

		/*
		 * while this struct direct is in this block, and
		 * there is still data to transfer in the user's request
		 */
		while ((dctp < (struct direct *)(bp->b_un.b_addr + on + n)) &&
			(tresid >= DNTSIZ(dctp->d_namlen)) &&
			(u.u_error == 0)) {

			if (len = dirmangled(ip, dctp, diroff)) {
				diroff += len;
				dctp = (struct direct *)
					((int)dctp + len);
#ifdef DEBUG
				getdents_debug[16]++;
#endif /* DEBUG */
				continue;
			}

			if (dctp->d_fileno == 0) {
				diroff += dctp->d_reclen;
				dctp = (struct direct *)((caddr_t)dctp +
						dctp->d_reclen);
#ifdef DEBUG
				getdents_debug[13]++;
				if (getdents_lvl)
					printf("empty entry to diroff %d\n",
						diroff);
#endif /* DEBUG */
				continue;
			}

#ifdef DEBUG
			getdents_debug[7]++;
#endif /* DEBUG */
			/*
			 * Is there room for this record in the
			 * translation buffer?  If no, move the data out
			 * and reset the pointer.  otherwise,
			 * do another translation.
			 */
			if (((caddr_t)dntp + DNTSIZ(dctp->d_namlen)) >=
					(ebp->b_un.b_addr + bsize)) {
				u.u_error =
		    		uiomove(ebp->b_un.b_addr,
					(caddr_t)dntp - ebp->b_un.b_addr,
					UIO_READ, uio);
				dntp = (struct dirent *)ebp->b_un.b_addr;
				if (u.u_error == 0)
					ASSERT(tresid == uio->uio_resid,
						"ufs_getdents: uio_resid sync");
					/*
					 *+ The algorithms for
					 *+ translating from the on-disk
					 *+ directory format to the filesystem-independent
					 *+ directory format have
					 *+ gotten out of synch.
					 */
#ifdef DEBUG
				getdents_debug[8]++;
				if (getdents_lvl)
					printf("uiomove1 %d\n", uio->uio_resid);
#endif /* DEBUG */
			} else {
				/*
				 * load with offset to NEXT entry (ugh)
				 */
				diroff += dctp->d_reclen;
				dntp->d_off = diroff;
				dntp->d_ino =  dctp->d_fileno;
				dntp->d_reclen =  DNTSIZ(dctp->d_namlen);
				/* copy name plus null pad */
				bcopy(dctp->d_name, dntp->d_name,
					(unsigned)dctp->d_namlen+1);
#ifdef DEBUG
				if (getdents_lvl)
					printf("pre-trans diroff %d tresid %d\n",
						diroff, tresid);
#endif /* DEBUG */
				tresid -= dntp->d_reclen;
				dntp = (struct dirent *)((caddr_t)dntp +
						dntp->d_reclen);
				dctp = (struct direct *)((caddr_t)dctp +
						dctp->d_reclen);
#ifdef DEBUG
				getdents_debug[9]++;
				if (getdents_lvl)
					printf("post_trans diroff %d tresid %d\n",
						diroff, tresid);
#endif /* DEBUG */
			}
		}
		if (n + on == bsize || diroff == ip->i_size) {
			bp->b_flags |= B_AGE;
#ifdef DEBUG
			getdents_debug[10]++;
#endif /* DEBUG */
		}
		brelse(bp);
	/*
	 * repeat loop as long as
	 *	there have been no errors
	 *	The user's buffer has at least space for 1 more entry.
	 *	We've exhausted the current buffer.
	 *	The last buffer we got had data in it to transfer.
	 */
	} while (u.u_error == 0 &&
		tresid >= DNTSIZ(1) &&
		(dctp >= (struct direct *)(bp->b_un.b_addr + on + n)) &&
		n != 0);

finish:
	/*
	 * If there's data in the buffer to transfer, and there's been
	 * no error, do so.
	 */
	if ((dntp != (struct dirent *)ebp->b_un.b_addr) && (u.u_error == 0)) {
		u.u_error = uiomove(ebp->b_un.b_addr,
			(caddr_t)dntp - ebp->b_un.b_addr, UIO_READ, uio);
#ifdef DEBUG
		getdents_debug[11]++;
		if (u.u_error == 0)
			ASSERT(tresid == uio->uio_resid,
				"ufs_getdents: uio_resid sync1");
			/*
			 *+ The algorithms for
			 *+ translating from bsd
			 *+ directory format to att
			 *+ directory format has
			 *+ gotten out of synch.
			 */
		if (getdents_lvl)
			printf("uiomove2 %d\n", uio->uio_resid);
#endif /* DEBUG */
	}
	brelse(ebp);
	if (error == 0)				/* XXX */
		error = u.u_error;		/* XXX */
	/*
	 * restore uio's offset to reflect offset in directory, 
	 * rather than offset in "translated" directory space.
	 */
	uio->uio_offset = diroff;
	return (error);
}

/*
 * ufs_symlink()
 *	Create a symbolic link.
 *
 * Caller passes dvp unlocked.  It remains unlocked at exit.
 */

/*ARGSUSED*/
static
ufs_symlink(dvp, lnm, vap, tnm, cred)
	struct vnode *dvp;
	char *lnm;
	struct vattr *vap;
	char *tnm;
	struct ucred *cred;		/* not used */
{
	struct inode *ip;
	register int error;

	ip = (struct inode *) 0;
	vap->va_type = VLNK;
	vap->va_rdev = 0;
	error = direnter(VTOI(dvp), lnm, DE_CREATE,
		(struct inode *)0, (struct inode *)0, vap, &ip, 0);
	if (error == 0) {
		/*
		 * Clear out the csymlink fields, in over-compensation
		 * for previous bug where they weren't cleared
		 * at csymlink deletion time.
		 */
		ip->i_pflags &= ~IP_CSYMLN;
		ip->i_ucbln = 0;
		ip->i_attln = 0;
		error = rdwri(UIO_WRITE, ip,
				tnm, strlen(tnm), 0, UIOSEG_KERNEL, (int *)0);
		IPUT(ip);
	} else if (error == EEXIST) {
		IPUT(ip);
	}
	return (error);
}

/*
 * rdwri()
 *	Read/write an inode.
 *
 * Caller passes ip locked.
 */

rdwri(rw, ip, base, len, offset, seg, aresid)
	enum uio_rw rw;
	struct inode *ip;
	caddr_t base;
	int len;
	int offset;
	int seg;
	int *aresid;
{
	struct uio auio;
	struct iovec aiov;
	register int error;

	aiov.iov_base = base;
	aiov.iov_len = len;
	auio.uio_iov = &aiov;
	auio.uio_iovcnt = 1;
	auio.uio_offset = offset;
	auio.uio_segflg = seg;
	auio.uio_resid = len;
	error = ufs_rdwr(ITOV(ip), &auio, rw, 0, u.u_cred);
	if (aresid) {
		*aresid = auio.uio_resid;
	} else if (auio.uio_resid) {
		error = EIO;
	}
	return (error);
}

/*
 * ufs_bmap()
 *	Return "physical" block number of an inode for a given logical
 *	block number.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

static
ufs_bmap(vp, lbn, vpp, bnp, rwflg, size)
	struct vnode *vp;
	daddr_t lbn;
	struct vnode **vpp;
	daddr_t *bnp;
	int rwflg;
	int size;
{
	register struct inode *ip;

	ip = VTOI(vp);
	if (vpp)
		*vpp = ip->i_devvp;
	if (bnp)
		*bnp = fsbtodb(ip->i_fs, bmap(ip, lbn, rwflg, size));
	return (0);
}

/*
 * ufs_bread()
 *	Read a logical block and return it in a buffer.
 *
 * Caller passes vp locked.  It remains locked at exit.
 */

static
ufs_bread(vp, lbn, bpp)
	struct vnode *vp;
	daddr_t lbn;
	struct buf **bpp;
{
	register struct inode *ip;
	register struct buf *bp;
	register daddr_t bn;
	register int size;

	ip = VTOI(vp);
	size = blksize(ip->i_fs, ip, lbn);
	bn = fsbtodb(ip->i_fs, bmap(ip, lbn, B_READ));
	if ((long)bn < 0) {
		bp = geteblk(size);
		clrbuf(bp);
	} else if (ip->i_lastr + 1 == lbn) {
		bp = breada(ip->i_devvp, bn, size, l.rablock, l.rasize);
	} else {
		bp = bread(ip->i_devvp, bn, size);
	}
	ip->i_lastr = lbn;
	IMARK(ip, IACC);
	if (bp->b_flags & B_ERROR) {
		brelse(bp);
		return (EIO);
	} else {
		*bpp = bp;
		return (0);
	}
}

/*
 * ufs_brelse()
 *	Release a block returned by ufs_bread.
 */

static
ufs_brelse(bp)
	struct buf *bp;
{
	bp->b_flags |= B_AGE;
	bp->b_resid = 0;
	brelse(bp);
}

/*
 * ufs_realvp()
 *	Return the actual vnode, if any, represented by this snode.
 *
 * Caller passes vp unlocked.  vp and vpp are unlocked at exit.
 *
 * Since there is no other vnode, returning any error will cause
 * the caller to use the original vnode (if appropriate).
 */

/*ARGSUSED*/
static int
ufs_realvp(vp, vpp)
	struct vnode *vp;
	struct vnode **vpp;
{
	return (EINVAL);
}

/*ARGSUSED*/
static int
ufs_nostrmsgioop(vp, mctl, mdata, rw, uflag, ioflag)
	struct vnode *vp;
	struct strbuf *mctl;
	struct strbuf *mdata;
	enum uio_rw rw;
	int uflag;
	int ioflag;
{
	return (ENOSTR);
}

/*ARGSUSED*/
static int
ufs_nostrpollop(vp, events, anyyet)
	struct vnode *vp;
	short events;
	int anyyet;
{
	return (POLLNVAL);
}

static
ufs_badop()
{
	cmn_err(CE_PANIC, "ufs_badop");
	/*
	 *+ This filesystem type does not support the requested operation.
	 */
}

/*
 * UNIX file-system vnode-operations interface structure.
 */

extern	int	iinactive();

struct vnodeops ufs_vnodeops = {
	ufs_open,	 /* vn_open     */
	ufs_close,	 /* vn_close    */
	ufs_rdwr,	 /* vn_rdwr     */
	ufs_ioctl,	 /* vn_ioctl    */
	ufs_getattr,	 /* vn_geattr   */
	ufs_setattr,	 /* vn_setattr  */
	ufs_access,	 /* vn_access   */
	ufs_lookup,	 /* vn_lookup   */
	ufs_create,	 /* vn_creat    */
	ufs_remove,	 /* vn_remove   */
	ufs_link,	 /* vn_link     */
	ufs_rename,	 /* vn_rename   */
	ufs_mkdir,	 /* vn_mkdir    */
	ufs_rmdir,	 /* vn_rmdir    */
	ufs_readdir,	 /* vn_readdir  */
	ufs_symlink,	 /* vn_symlink  */
	ufs_readlink,	 /* vn_readlink */
	ufs_fsync,	 /* vn_fsync    */
	iinactive,	 /* vn_inactive */
	ufs_bmap,	 /* vn_bmap     */
	ufs_badop,	 /* vn_strategy */
	ufs_bread,	 /* vn_bread    */
	ufs_brelse,	 /* vn_brelse   */
	ufs_badop,	 /* vn_minphys  */
	ufs_getdents,	 /* vn_getdents */
	ufs_fdclose,	 /* vn_fdclose  */
	ufs_fcntl,	 /* vn_fcntl    */
	ufs_realvp,	 /* vn_realvp   */
	ufs_badop,	 /* vn_vhangup  */
	ufs_nostrmsgioop,/* vn_strmsgio */
	ufs_nostrpollop, /* vn_strpoll  */
};
